__author__ = "Tommaso Barbugli"
__copyright__ = "Copyright 2019-2021, Stream.io, Inc"
__version__ = "3.12.0"
__maintainer__ = "Tommaso Barbugli"
__email__ = "support@getstream.io"
__status__ = "Production"
